unit UMachine;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type
  TOpcode = (opNop, opHalt, opLD, opLDC, opLDM, opSTM, opInc, opDec, opJP, opJr,
    opPush, opPop, opCall, opRet, opAdd, opSub, opMul, opDiv, opShr, opShl,
    opIn, opOut, opXchg, opXor, opOr, opAnd, opJpc, opJrc, opTstZ, opTstNZ,
    opLDMD, opSTMD);

  TOpcodeHandler = procedure of object;

  { TCPU }

  generic TCPU<T> = class
  type
    TOutputEvent = procedure (Port: T; Value: T) of object;
    TInputEvent = function (Port: T): T of object;
  var
    Terminated: Boolean;
    OpcodeHandlers: array[TOpcode] of TOpcodeHandler;
    function ReadNext: T;
    procedure Step;
    procedure AddOpcode(Opcode: TOpcode; Handler: TOpcodeHandler);
    procedure InitOpcodes;
  private
    FOnInput: TInputEvent;
    FOnOutput: TOutputEvent;
    Condition: Boolean;
    procedure OpcodeAddition;
    procedure OpcodeAnd;
    procedure OpcodeCall;
    procedure OpcodeDecrement;
    procedure OpcodeDivision;
    procedure OpcodeExchange;
    procedure OpcodeHalt;
    procedure OpcodeIncrement;
    procedure OpcodeInput;
    procedure OpcodeJump;
    procedure OpcodeJumpConditional;
    procedure OpcodeJumpRelative;
    procedure OpcodeJumpRelativeConditional;
    procedure OpcodeLoad;
    procedure OpcodeLoadConst;
    procedure OpcodeLoadMemory;
    procedure OpcodeLoadMemoryDisplacement;
    procedure OpcodeMultiplication;
    procedure OpcodeNop;
    procedure OpcodeOr;
    procedure OpcodeOutput;
    procedure OpcodePop;
    procedure OpcodePush;
    procedure OpcodeReturn;
    procedure OpcodeShiftLeft;
    procedure OpcodeShiftRight;
    procedure OpcodeStoreMemory;
    procedure OpcodeStoreMemoryDisplacement;
    procedure OpcodeSubtraction;
    procedure OpcodeTestNotZero;
    procedure OpcodeTestZero;
    procedure OpcodeXor;
  public
    Registers: array of T;
    Memory: array of T;
    IP: T;
    SP: T;
    procedure Run;
    constructor Create;
    property OnInput: TInputEvent read FOnInput write FOnInput;
    property OnOutput: TOutputEvent read FOnOutput write FOnOutput;
  end;


implementation

{ TCPU }

function TCPU.ReadNext: T;
begin
  Result := Memory[IP];
  Inc(IP);
end;

procedure TCPU.Run;
begin
  IP := 0;
  SP := Length(Memory);
  Terminated := False;
  while not Terminated do
    Step;
end;

constructor TCPU.Create;
begin
  SetLength(Registers, 32);
  SetLength(Memory, 1000);
  InitOpcodes;
end;

procedure TCPU.OpcodeNop;
begin
end;

procedure TCPU.OpcodeHalt;
begin
  Terminated := True;
end;

procedure TCPU.OpcodeLoad;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Src];
end;

procedure TCPU.OpcodeLoadConst;
var
  Reg: T;
begin
  Reg := ReadNext;
  Registers[Reg] := ReadNext;
end;

procedure TCPU.OpcodeLoadMemory;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Memory[Registers[Src]];
end;

procedure TCPU.OpcodeStoreMemory;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Memory[Registers[Dest]] := Registers[Src];
end;

procedure TCPU.OpcodeLoadMemoryDisplacement;
var
  Reg, Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Reg := ReadNext;
  Registers[Dest] := Memory[Registers[Src] + Registers[Reg]];
end;

procedure TCPU.OpcodeStoreMemoryDisplacement;
var
  Reg, Src, Dest: T;
begin
  Dest := ReadNext;
  Reg := ReadNext;
  Src := ReadNext;
  Memory[Registers[Dest] + Registers[Reg]] := Registers[Src];
end;

procedure TCPU.OpcodeIncrement;
var
  Reg: T;
begin
  Reg := ReadNext;
  Registers[Reg] := Registers[Reg] + 1;
end;

procedure TCPU.OpcodeDecrement;
var
  Reg: T;
begin
  Reg := ReadNext;
  Registers[Reg] := Registers[Reg] - 1;
end;

procedure TCPU.OpcodeAddition;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] + Registers[Src];
end;

procedure TCPU.OpcodeSubtraction;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] - Registers[Src];
end;

procedure TCPU.OpcodeMultiplication;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] * Registers[Src];
end;

procedure TCPU.OpcodeDivision;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] div Registers[Src];
end;

procedure TCPU.OpcodeShiftLeft;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] shl Registers[Src];
end;

procedure TCPU.OpcodeShiftRight;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] shr Registers[Src];
end;

procedure TCPU.OpcodeAnd;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] and Registers[Src];
end;

procedure TCPU.OpcodeOr;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] or Registers[Src];
end;

procedure TCPU.OpcodeXor;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Registers[Dest] := Registers[Dest] xor Registers[Src];
end;

procedure TCPU.OpcodeJump;
var
  Reg: T;
begin
  Reg := ReadNext;
  IP := Registers[Reg];
end;

procedure TCPU.OpcodeJumpConditional;
var
  Reg: T;
begin
  Reg := ReadNext;
  if Condition then
    IP := Registers[Reg];
end;

procedure TCPU.OpcodeJumpRelative;
var
  Reg: T;
begin
  Reg := ReadNext;
  IP := IP + Registers[Reg];
end;

procedure TCPU.OpcodeJumpRelativeConditional;
var
  Reg: T;
begin
  Reg := ReadNext;
  if Condition then
    IP := IP + Registers[Reg];
end;

procedure TCPU.OpcodePush;
var
  Reg: T;
begin
  SP := SP - 1;
  Reg := ReadNext;
  Memory[SP] := Registers[Reg];
end;

procedure TCPU.OpcodePop;
var
  Reg: T;
begin
  Reg := ReadNext;
  Registers[Reg] := Memory[SP];
  SP := SP + 1;
end;

procedure TCPU.OpcodeCall;
var
  Reg: T;
begin
  SP := SP - 1;
  Reg := ReadNext;
  Memory[SP] := IP;
  IP := Registers[Reg];
end;

procedure TCPU.OpcodeReturn;
begin
  IP := Memory[SP];
  SP := SP + 1;
end;

procedure TCPU.OpcodeInput;
var
  Src, Dest: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  if Assigned(FOnInput) then
    Registers[Dest] := FOnInput(Registers[Src]);
end;

procedure TCPU.OpcodeOutput;
var
  Dest, Src: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  if Assigned(FOnOutput) then
    FOnOutput(Registers[Dest], Registers[Src]);
end;

procedure TCPU.OpcodeTestZero;
var
  Reg: T;
begin
  Reg := ReadNext;
  Condition := Registers[Reg] = 0;
end;

procedure TCPU.OpcodeTestNotZero;
var
  Reg: T;
begin
  Reg := ReadNext;
  Condition := Registers[Reg] <> 0;
end;

procedure TCPU.OpcodeExchange;
var
  Dest, Src, Temp: T;
begin
  Dest := ReadNext;
  Src := ReadNext;
  Temp := Registers[Dest];
  Registers[Dest] := Registers[Src];
  Registers[Src] := Temp;
end;


procedure TCPU.Step;
var
  Opcode: TOpcode;
  Handler: TOpcodeHandler;
begin
  Opcode := TOpcode(ReadNext);
  if Opcode <= High(TOpcode) then begin
    Handler := OpcodeHandlers[Opcode];
    if Assigned(Handler) then Handler
    else raise Exception.Create('Missing handler for opcode + ' + IntToStr(Integer(Opcode)));
  end else raise Exception.Create('Unknown opcode: ' + IntToStr(Integer(Opcode)));
end;

procedure TCPU.AddOpcode(Opcode: TOpcode; Handler: TOpcodeHandler);
begin
  OpcodeHandlers[Opcode] := Handler;
end;

procedure TCPU.InitOpcodes;
begin
  AddOpcode(opNop, @OpcodeNop);
  AddOpcode(opHalt, @OpcodeHalt);
  AddOpcode(opAnd, @OpcodeAnd);
  AddOpcode(opOr, @OpcodeOr);
  AddOpcode(opXor, @OpcodeXor);
  AddOpcode(opLD, @OpcodeLoad);
  AddOpcode(opLDC, @OpcodeLoadConst);
  AddOpcode(opLDM, @OpcodeLoadMemory);
  AddOpcode(opSTM, @OpcodeStoreMemory);
  AddOpcode(opLDMD, @OpcodeLoadMemoryDisplacement);
  AddOpcode(opSTMD, @OpcodeStoreMemoryDisplacement);
  AddOpcode(opCall, @OpcodeCall);
  AddOpcode(opRet, @OpcodeReturn);
  AddOpcode(opPush, @OpcodePush);
  AddOpcode(opPop, @OpcodePop);
  AddOpcode(opJP, @OpcodeJump);
  AddOpcode(opJpc, @OpcodeJumpConditional);
  AddOpcode(opJR, @OpcodeJumpRelative);
  AddOpcode(opJRC, @OpcodeJumpRelativeConditional);
  AddOpcode(opAdd, @OpcodeAddition);
  AddOpcode(opSub, @OpcodeSubtraction);
  AddOpcode(opMul, @OpcodeMultiplication);
  AddOpcode(opDiv, @OpcodeDivision);
  AddOpcode(opXchg, @OpcodeExchange);
  AddOpcode(opShl, @OpcodeShiftLeft);
  AddOpcode(opShr, @OpcodeShiftRight);
  AddOpcode(opIn, @OpcodeInput);
  AddOpcode(opOut, @OpcodeOutput);
  AddOpcode(opInc, @OpcodeIncrement);
  AddOpcode(opDec, @OpcodeDecrement);
  AddOpcode(opTstNZ, @OpcodeTestNotZero);
  AddOpcode(opTstZ, @OpcodeTestZero);
end;

end.

